(* Time-stamp: <modified the 30/05/2022 (at 14:49) by Erwan Jahier> *)

open Graph
open Graph.Dot_ast

type node_id = string
type node = {
  id: node_id;
  file: string;
  init: string;
}


type t = {
  nodes: node list;
  pred: node_id -> (int * node_id) list;
  succ: node_id -> node_id list;
  of_id: node_id -> node;
  directed:bool;
  attributes:  (string * string) list;
}

type node_info_t = (string, node) Hashtbl.t
let node_info:node_info_t = Hashtbl.create 100

type node_pred_t = (string, int * node_id) Hashtbl.t
let node_pred:node_pred_t = Hashtbl.create 100
type node_succ_t = (string, node_id) Hashtbl.t
let node_succ:node_succ_t = Hashtbl.create 100

let clean_tbl () =
  Hashtbl.clear node_info;
  Hashtbl.clear node_pred;
  Hashtbl.clear node_succ

let (of_id:Dot_ast.id -> string) =
  function Ident str | Html str | Number str | String str -> str

let (of_node_id:Dot_ast.node_id -> string) = fun id -> of_id (fst id)
let (of_node:Dot_ast.node -> string) = function
  | NodeId node_id -> of_node_id node_id
  | NodeSub _subgraph -> failwith "subgraphs are not supported in sasa "


let (get_file: Dot_ast.node_id  -> Dot_ast.attr list -> string) =
  fun node_id attrs ->
    let attrs = List.flatten attrs in (* XXX why a list of list ? *)
    try  (
        match List.assoc (Ident "algo") attrs with
          None -> assert false
        | Some id -> of_id id
      )
    with Not_found -> failwith ((of_node_id node_id)^" should have an algo attribute")

let (get_init: Dot_ast.attr list -> string) =
  fun attrs ->
    let attrs = List.flatten attrs in (* XXX why a list of list ? *)
    let init_list =
      List.fold_left
        (fun acc (id,idopt) ->
           if id <> Ident "init" then acc else
             match idopt with
             | Some (String id) -> id
             | _ -> acc
        )
        ""
        attrs
    in
    init_list

let compare_alpha_num str1 str2 =
  (* if str1 and str2 are of the form [^0-9][0-9]+, we use the alphabetic
     order for the first part, and the numeric order for the second part. 
     nb : this is the case for node names generated by gg
  *)
    try
      let to_int s = int_of_string (String.sub s 1 (String.length s - 1)) in
      compare (str1.[0], to_int str1) (str2.[0], to_int str2)
    with
      _ -> String.compare str1 str2

let compare_alpha_num_2 p1 p2 = compare_alpha_num (snd p1) (snd p2)

let rec (get_weight: Dot_ast.attr -> int) =
  function
  | (Ident "weight", Some Number n)::_
  | (Ident "weight", Some String n)::_ -> (try (int_of_string n) with _  -> 1)
  | _::tail -> get_weight tail
  | [] -> 1

type attrs = (string * string) list

let (do_graph_attr : attrs -> (Dot_ast.id * Dot_ast.id option) list -> attrs) =
  fun acc l ->
    let f acc = function
      | id, Some v ->  (of_id id, of_id v)::acc
      | _, None -> acc
    in
    List.fold_left f acc l


let (do_stmt: bool -> node list * attrs -> Dot_ast.stmt -> node list * attrs) =
  fun directed (n, attrs) stmt ->
    match stmt with
    | Node_stmt (node_id, dot_attrs) ->
      let id = of_node_id node_id in
      let inits = get_init dot_attrs  in
      let node = { id=id ; file = get_file node_id dot_attrs ; init = inits } in
      Hashtbl.add node_info id node;
      node::n, attrs
    | Edge_stmt (dot_node, nodes, dot_attrs) ->
      let node = of_node dot_node in
      let nodes = List.map of_node nodes in
        (* for egdes written "a -- b -- c -- d", which
           is a shortcut for "
           a -- b
           b -- c
           c -- d
           "
           Graph.Dot.parse_dot_ast returns the pair a,[b,c,d]
           which is weird IMHO.
           The code below add the missing edges:
        *)
      let dot_attrs = List.flatten dot_attrs in (* XXX why a list of list ? *)
      let weight = get_weight dot_attrs in
      let add_edge n1 n2 =
        if n1 = n2 then failwith
            (Printf.sprintf
               "Bad topology: %s can not ne a neighbor of itself!" n1);
        let pn1 = Hashtbl.find_all node_pred n1 in
        let pn2 = Hashtbl.find_all node_pred n2 in
        if not (List.mem (weight,n1) pn2) then
          Hashtbl.add node_pred n2 (weight,n1);
        if not directed && not (List.mem (weight,n2) pn1) then
          Hashtbl.add node_pred n1 (weight,n2);
        n2
      in
      ignore (List.fold_left add_edge node nodes);
      n, attrs
    | Attr_graph dot_attrs ->
      let attrs = do_graph_attr attrs (List.flatten dot_attrs) in
      n, attrs
    | Attr_node  _attrs -> n, attrs
    | Attr_edge  _attrs ->  n, attrs
    | Equal (_id1, _id2) -> assert false
    | Subgraph _subgraph -> assert false


let (read: string -> t) = fun f ->
  clean_tbl ();
  let dot_file = Graph.Dot.parse_dot_ast f in
  assert (not dot_file.strict);
  let nodes, attrs =
    List.fold_left (do_stmt dot_file.digraph) ([], []) dot_file.stmts
  in
  Hashtbl.iter
    (fun pid (_, pid_pred) ->
       Hashtbl.add node_succ pid pid_pred
    )
    node_pred;
  let succ str = List.sort compare_alpha_num (Hashtbl.find_all node_succ str) in
  let pred str = List.sort compare_alpha_num_2 (Hashtbl.find_all node_pred str) in
  {
    nodes = List.rev nodes;
    succ = succ;
    pred = pred;
    of_id = (fun str -> try Hashtbl.find node_info str with Not_found ->
                          failwith (str^ " unknown node id")
            );
    directed = dot_file.digraph;
    attributes = List.rev attrs
  }

let (to_adjacency: t -> bool array array) =
  fun t ->
    let n = List.length t.nodes in
    let rank_node_tbl = Hashtbl.create n in
    let m = Array.make_matrix n n false in
    let rank_node = Hashtbl.find rank_node_tbl in
    List.iteri (fun i n -> Hashtbl.add rank_node_tbl n.id i) t.nodes;
    List.iteri
      (fun i n  ->
         List.iter
           (fun (target) -> m.(i).(rank_node target) <- true)
           (t.succ n.id)
      )
      t.nodes;
    m

let (get_degree: t -> int*int) =
  fun t ->
  match t.nodes with
  | [] -> 0,0
  | n::tl ->
    let node_deg n =
      if t.directed then
        List.length (t.succ (n.id)) + List.length (t.pred (n.id))
      else
        List.length (t.succ (n.id))
    in
    let d0 = node_deg n in
    let dmin, dmax = List.fold_left
        (fun (d_min,d_max) n ->
           (min (node_deg n) d_min, max (node_deg n) d_max)
        )
        (d0, d0)
        tl
    in
    dmin, dmax

let (get_nb_link: t -> int) =
  fun t ->
  let res = List.fold_left
      (fun acc n ->
         (* let succ = String.concat "," (List.map snd (t.succ n.id)) in *)
         (* Printf.printf "%s->%s\n%!" n.id succ; *)
         ((List.length (t.succ n.id))) + acc) 0 t.nodes
  in
  if t.directed then res else res/2

let (get_mean_degree : t -> float) =
  fun t ->
    2.0 *. (float_of_int (get_nb_link t)) /. (float_of_int (List.length t.nodes))

type color = W | G | B
exception Cycle
let directed_is_cyclic : t -> bool =
  fun g ->
  assert (g.directed);
  let t = Hashtbl.create (List.length g.nodes) in
  let nodes = List.map (fun n -> 0 (* fake weight *), n.id) g.nodes in
  let color pid = match Hashtbl.find_opt t pid with
      Some c -> c | None -> assert false
  in
  List.iter (fun (_, n) -> Hashtbl.add t n W) nodes;
  let rec visit (_,pid) =
    match color pid with
    | G -> raise Cycle
    | B -> ()
    | W ->
      Hashtbl.replace t pid G;
      List.iter visit (g.pred pid);
      Hashtbl.replace t pid B
  in
  try List.iter visit nodes; false
  with Cycle -> true

let is_connected : t -> bool =
  fun g ->
  let visited = Hashtbl.create 0 in
  let rec f acc pid =
    if Hashtbl.mem visited pid then acc else
      (Hashtbl.add visited pid pid;
       List.fold_left f (pid::acc)
         (List.rev_append (g.succ pid) (List.map snd (g.pred pid)) )
      )
  in
  let accessible = f [] (List.hd g.nodes).id in
  List.length accessible = List.length g.nodes

let bfs : (t -> string -> bool * string list) =
  fun t n ->
  let q = Queue.create () in
  let discovered = ref [n] and parent = ref (function _ -> "") in
  let cyclic = ref false in
  Queue.add n q;
  while not (Queue.is_empty q) do
    let node = Queue.take q in
    parent :=
      List.fold_left
        (fun parents suc ->
           if List.for_all (fun disc -> disc <> suc) !discovered
           then (
             Queue.add suc q;
             discovered := (suc)::!discovered;
             function a -> if a = suc then node else parents a
           ) else ((
               if suc <> (parents node)
               then
                 cyclic := true);
              parents
             ))
        !parent
        (t.succ node)
  done;
  (!cyclic, !discovered)

let non_directed_is_cyclic : t -> bool =
  fun t ->
  assert(not t.directed);
  match t.nodes with
    | [] -> false
    | hd::_ ->
      let (cyclic, _bfs_nodes) = (bfs t hd.id) in
      cyclic

let is_cyclic : t -> bool =
  fun g ->
  if g.directed then directed_is_cyclic g else non_directed_is_cyclic g

let is_tree : t -> bool =
  fun g ->
  (not (is_cyclic g)) && (is_connected g)

let is_root_pid pid = Str.string_match (Str.regexp ".*[rR]oot.*") pid 0
let is_root_node n = is_root_pid n.id

(* cf algo.mli for the spec *)
let is_rooted_tree : t -> bool =
  fun g ->
  is_tree g && List.length (List.filter is_root_node g.nodes) = 1

(** Returns the channel number that let [p_neighbor] access to the
   content of [p], if [p] is a neighbor of [p_neighbor].  Returns -1 if
   [p] is not a neighbor of [p_neighbor], which can happen in directed
   graphs. *)
let (reply: t -> string -> string -> int) =
  fun g p p_neighbor  ->
    let rec f i = function
      | [] -> (-1) (* may happen in directed graphs *)
      | (_,x)::t -> if x=p then i else f (i+1) t
    in
    f 0 (g.pred p_neighbor)

let (reply_succ: t -> string -> string -> int) =
  fun g p p_neighbor  ->
    let rec f i = function
      | [] -> (-1) (* may happen in directed graphs *)
      | x::t -> if x=p then i else f (i+1) t
    in
    f 0 (g.succ p_neighbor)

(* count the occurrences of x in l *)
let _count x l =
  let rec f acc = function
    | [] -> acc
    | y::t -> f (if x=y then acc+1 else acc) t
  in
  f 0 l

let is_in_tree g =
  g.directed && (* redundant, but more efficient *)
  (is_tree g) &&
  (List.for_all (fun n -> List.length (g.pred n.id) <= 1) g.nodes)

let is_out_tree g =
  g.directed && (* redundant, but more efficient *)
  (is_tree g) &&
  (List.for_all (fun n -> List.length (g.succ n.id)  <= 1) g.nodes)


(* Donne les enfants d'un noeud dans un in-tree (liens partent de la racine) *)
let children_in g pid =
  g.succ pid

(* Donne les enfants d'un noeud dans un out-tree (liens vers la racine) *)
let children_out g pid =
  List.map snd (g.pred pid)

(* Donne les enfants d'un noeud dans un in-out-tree *)
let children_in_out g pid =
  assert (is_rooted_tree g);
  let succ = g.succ pid in
  if is_root_pid pid then succ
  else List.tl succ
  (* pour tous les noeuds sauf la racine, le parent est la tête de succ
  ici on veut juste les enfants donc on enlève le premier *)

let parent_in g pid =
  match g.pred pid with
  | [] -> None
  | (_,id)::_ -> Some id

let parent_out g pid =
  match g.succ pid with
  | [] -> None
  | id::_ -> Some id

let parent_in_out g pid =
  assert (is_rooted_tree g);
  if is_root_pid pid then None
  else Some (List.hd (g.succ pid))
  (* Le parent est le premier dans la liste succ pour un in-out-tree ou un rooted-tree *)

let get_parent = fun g pid ->
  assert (is_tree g);
  let in_tree = is_in_tree g in
  let out_tree = is_out_tree g in
  let get_parent, rep =
    if out_tree then
      parent_out, reply
    else if in_tree then
      parent_in, reply_succ
    else
      parent_in_out, reply
  in
  Option.map (fun par -> rep g par pid) (get_parent g pid)
  (* on renvoie le numéro de canal du parent *)

(* Calcule la hauteur d'un noeud.
   Hauteur(feuille) = 0
   Hauteur(noeud) = 1 + max(hauteurs des enfants)
*)
let get_height g pid: int =
  assert (is_tree g);
  let in_tree = is_in_tree g in
  let out_tree = is_out_tree g in
  let get_children =
    if in_tree then children_in
    else if out_tree then children_out
    else children_in_out
  in
  let rec height children g pid =
    match children g pid with
    | [] -> 0
    | x::xs ->
      let hx = height children g x in
      let hxs = List.map (height children g) xs in
      1 + List.fold_left max hx hxs
  in
  height get_children g pid

(* Calcule le niveau/la profondeur d'un noeud
   Niveau(racine) = 0
   Niveau(noeud) = Niveau(noeud.parent) + 1
*)
let get_level g pid: int =
  assert (is_tree g);
  let in_tree = is_in_tree g in
  let out_tree = is_out_tree g in
  let get_parent =
    if in_tree then parent_in
    else if out_tree then parent_out
    else parent_in_out
  in
  let rec level parent g pid =
    match parent g pid with
    | None -> 0
    | Some par -> 1 + (level parent g par)
  in
  level get_parent g pid

(* Calcule la taille du sous-arbre partant du noeud donné
   taille = 1 + somme des tailles des enfants
*)
let get_subtree_size g pid: int =
  assert (is_tree g);
  let in_tree = is_in_tree g in
  let out_tree = is_out_tree g in
  let get_children =
    if in_tree then children_in
    else if out_tree then children_out
    else children_in_out
  in
  let rec tree_size children g pid =
    1 + (match children g pid with
      | [] -> 0
      | l -> List.fold_left (+) 0 (List.map (tree_size children g) l)
    )
  in
  tree_size get_children g pid
