(* This file defines sasa specific commands. 

It is meant  to be loaded after the (rdbg-generated) "rdbg-cmds.ml" file.
*)
open RdbgMain
#require "sasacore";;

open Sasacore.Topology;;
#use "dot4sasa.ml";;



let is_silent ?(dflt=true) e =
  match List.assoc_opt "silent" e.data with
  | Some B b -> b
  | _ ->
     Printf.printf "The silent value is not available at event %d\n%!" e.nb;
     dflt

let is_legitimate e =
  match List.assoc_opt "legitimate" e.data with
  | Some B b -> b
  | _ -> failwith ("legitimate not available at this event: "^(string_of_event e))

(**********************************************************************)
(** Computing rounds *)

module IntMap = Map.Make(Int)

type round_st = {
    cpt: int;
    mask : string list;  (* nodes we look the activation of *) (* XXX use an array! *)
    tbl : (int * bool * string list) IntMap.t
                                     (* maps event nb to round, round nb, and mask *)
  }

let verbose = ref false

let round_st_init = {
    cpt = 1;
    mask = [];
    tbl = IntMap.empty
  }

let round_st_ref = ref round_st_init

let set_round_st_cpt n = round_st_ref := { !round_st_ref with cpt = n }
let set_round_st_mask m = round_st_ref := { !round_st_ref with mask = m }
let set_round_st_tbl t = round_st_ref := { !round_st_ref with tbl = t }

let round_reset i =
  round_st_ref := round_st_init;
  if !verbose then Printf.printf "\nInit round_st at event %d\n%!" i;
  set_round_st_tbl (IntMap.add i (1,true, []) !round_st_ref.tbl)

let clean_round_st_tbl i =
  set_round_st_tbl (IntMap.filter (fun k _v -> k <= i) !round_st_ref.tbl) 
                       
(* a process can be removed from the mask if one action of p is triggered
  or if no action of p is enabled *)
let get_removable pl =
  let pl = List.filter
      (fun p ->
         (List.exists (fun (_,_,acti) -> acti) p.actions) ||
         (List.for_all (fun (_,enab,_) -> (not enab)) p.actions)
      )
      pl
  in
  List.map (fun p -> p.name) pl

let enabled pl = (* returns the enabled processes *)
  let el = List.filter
    (fun p -> List.exists (fun (_,enab,_) -> enab) p.actions)
    pl
  in
  List.map (fun p -> p.name) el

(* called at each event via the time-travel hook *)
let (round : RdbgEvent.t -> bool) =
  fun e ->
  match IntMap.find_opt e.nb !round_st_ref.tbl with
  | Some (croundnb, round, cmask) ->
     (*     Printf.printf "round tabulated at e.nb %d: croundnb=%d round = %b\n%!" *)
     (*       e.nb croundnb round; *)
     set_round_st_cpt croundnb;
     set_round_st_mask cmask;
     round
  | None ->
     if !round_st_ref.mask = [] && (e.data=[] || is_silent e) then false else
       let round =
         if not 
              ( (* we check if a round occurs when activated processes are available *)
                if args.salut_mode then
                  e.kind = Exit && e.name = "mv_hook" && e.step > 1
                else
                  e.kind = Ltop
              )
         then false
         else
           let (pl : process list) = get_processes e in
           if !round_st_ref.mask = [] then 
             (*  occurs  at the  first  possible  round, or  when  all
                processes in the previous mask were the only triggered
                processes *)
             set_round_st_mask (enabled pl);
           let rm_me = get_removable pl in
           if !verbose then (
             Printf.printf "\nMask (event %d): %s\n" e.nb
               (String.concat "," !round_st_ref.mask);
             Printf.printf "To remove from mask: %s\n%!" (String.concat "," rm_me)
           );
           set_round_st_mask
             (List.filter (fun pid -> not (List.mem pid rm_me)) !round_st_ref.mask);
           if !verbose then
             Printf.printf "New Mask: %s\n%!" (String.concat "," !round_st_ref.mask);
           let res = !round_st_ref.mask = [] in
           if res then (
             set_round_st_mask (
                 let mask =
                   List.filter
                     (fun p -> List.exists (fun (_,e,a) -> e && not(a)) p.actions)
                     pl
                 in
                 let mask = List.map (fun p -> p.name) mask in
                 if !verbose then (
                   let mask = List.rev mask in
                   Printf.printf "Next mask : %s\n%!" (String.concat "," mask);
                   flush stdout
                 );
                 mask
               )
           );
           res
       in
       if round then set_round_st_cpt (!round_st_ref.cpt + 1);
       if !verbose then
         Printf.printf "IntMap.add e.nb=%d round_nb=%d round= %b mask=%s\n%!"
           e.nb !round_st_ref.cpt round  (String.concat "," !round_st_ref.mask);
       set_round_st_tbl
         (IntMap.add e.nb (!round_st_ref.cpt, round, !round_st_ref.mask)
            !round_st_ref.tbl);
       (* Printf.printf "round computed at e.nb %d: croundnb=%d round = %b\n%!"
       e.nb !roundnb round; *)
       round

       
let restore_round_nb i =
  match IntMap.find_opt i !round_st_ref.tbl with
  | None ->
     (*  Printf.printf "restore_round_nb: At event %d, nothing to restore\n%!" e.nb ; *)
     ()
  | Some (n,_, m) ->
     (* Printf.printf "restore_round_nb: At event %d, restore the rn %d\n%!" e.nb n; *)
     set_round_st_cpt n;
     set_round_st_mask m;
     ()

let sasa_next e =
  let ne = e.next () in
  List.iter
    (fun str -> if str<>"print_event" then (RdbgStdLib.get_hook str) ne)
    (RdbgStdLib.list_hooks ()); 
  ne

let next_round e =
  next_cond_gen e round sasa_next

let back_step e =
  let e = rev_cond_gen e (fun ne -> ne.kind = e.kind && ne.name = e.name)
    sasa_next restore_round_nb
  in
  store e.nb;
  clean_round_st_tbl e.nb;
  e
  
(**********************************************************************)
(* redefine (more meaningful) step and back-step for sasa *)
let sasa_step e = next_cond e (fun ne -> ne.kind = e.kind && ne.name = e.name)
let s () = e:=sasa_step !e ; emacs_udate !e; store !e.nb;pe()
let b () =
  let ne = back_step !e in
  e:=ne ;
  emacs_udate !e; store !e.nb;pe()

let p =
  try Topology.read dotfile
  with _ -> failwith "This is not a sasa rdbg session!";;

let d_par () = dot true !round_st_ref.cpt p dotfile !e;;
let dot_par () = dot true !round_st_ref.cpt p dotfile !e;;
let ne_par () = neato true !round_st_ref.cpt p dotfile !e;;
let tw_par () = twopi true !round_st_ref.cpt p dotfile !e;;
let ci_par () = circo true !round_st_ref.cpt p dotfile !e;;
let fd_par () = fdp true !round_st_ref.cpt p dotfile !e;;
let sf_par () = sfdp true !round_st_ref.cpt p dotfile !e;;
let pa_par () = patchwork true !round_st_ref.cpt p dotfile !e;;
let os_par () = osage true !round_st_ref.cpt p dotfile !e;;

let dot () = dot false !round_st_ref.cpt p dotfile !e;;
let ne () = neato false !round_st_ref.cpt p dotfile !e;;
let tw () = twopi false !round_st_ref.cpt p dotfile !e;;
let ci () = circo false !round_st_ref.cpt p dotfile !e;;
let fd () = fdp false !round_st_ref.cpt p dotfile !e;;
let sf () = sfdp false !round_st_ref.cpt p dotfile !e;;
let pa () = patchwork false !round_st_ref.cpt p dotfile !e;;
let os () = osage false !round_st_ref.cpt p dotfile !e;;


(* To change the default dot/graph viewer: 
   dot_view := ci;;
 *)
let dot_view : (unit -> unit) ref =
  ref ( (* Try to chose the most sensible default viewer *)
      if String.length dotfile > 4 && String.sub dotfile 0 4 = "ring" then ci else
        if Algo.is_directed () then dot else ne)
  
let d () = !dot_view ()
let sd () = s();!dot_view();;
let bd()= e:=prev !e ; emacs_udate !e; pe();!dot_view();;


let nr () =
  let ne = next_round !e in
  e := ne;
  store !e.nb; !dot_view ();;

let pr () =
  e:=goto_last_ckpt !e.nb;
  restore_round_nb !e.nb;
  clean_round_st_tbl !e.nb;
  !dot_view ();
  store !e.nb
  
(* I need to overrides those *)

(* won't work in semi-auto modes, but the buttons are hided *)
let u () = undo (); ignore (round !e);; 
let r () =
  let seed = Seed.get dotfile in
  Seed.set seed;
  Printf.printf "Restarting using the seed %d" seed;
  !e.RdbgEvent.reset();
  e:=RdbgStdLib.run ~call_hooks:true ();
  round_reset !e.nb;
  redos := [1];
  (* ignore (round !e); *)
  (* if the first event is not a round, add it as a check_point *)
  (* if !ckpt_list = [] then *)
  ckpt_list := [!e];;

(**********************************************************************)
(* Move forward until silence *)

let goto_silence e = next_cond e is_silent 
let silence () = e:=goto_silence !e; !dot_view ();;

let _ = add_doc_entry
          "silence" "unit -> unit" "Move forward until is_silent returns true"
          "sasa" "sasa-rdbg-cmds.ml";
        add_doc_entry
          "is_silent" "RdbgEvent.t -> bool"
          "does the event correspond to a silent configuration? (i.e., no enable node)"
          "sasa" "sasa-rdbg-cmds.ml";;


let goto_legitimate e = next_cond e is_legitimate 
let legitimate () =  e:=goto_legitimate !e; !dot_view ();;
let _ =
  add_doc_entry
    "legitimate" "unit -> unit"
    "  Move forward until a legitimate configuration is reached (uses 'silence' by default)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry
    "is_legitimate" "RdbgEvent.t -> bool"
    "does the event correspond to a legitimate configuration?"
    "sasa" "sasa-rdbg-cmds.ml";;

(**********************************************************************)
(* print_event tuning *)

(* split  the vars into returns  (the enab vars, the  activated vars,
   the other vars) nb: in the "Enab" prefix is removed from enab vars
   names; ie we leave only the pid and the action name *)
type s = (string * string * Data.v) 
let split_data (l:Data.subst list) : s list * s list * s list = 
  let l = List.map (fun (x,v) -> Str.split (Str.regexp "_") x, v) l in
  let rec sortv (enab, other) (x,v) =
    match x with 
    | "Enab"::pid::tail -> (pid, String.concat "_" tail,v)::enab, other
    | pid::tail -> enab, (pid,(String.concat "_" tail),v)::other
    | [] -> assert false
  in
  let enab, other = List.fold_left sortv ([],[]) l in
  let acti_names = List.map (fun (pid, n, v) -> pid, n) enab in  
  let act, vars =
    List.partition (fun (pid,n, _) -> List.mem (pid,n) acti_names) other
  in
  enab, act, vars
  
  
let only_true l = List.filter (fun (_,_, v)  -> v = B true) l
(* Only print the active process values *)
let str_of_sasa_event short e =
  let enab, act, vars = split_data e.data in
  let enab = only_true enab in
  let act = only_true act in
  let act_pid = List.map (fun (pid,_,_) -> pid) act in
  let vars = List.filter (fun (pid, _,_) -> List.mem pid act_pid) vars in
  let _to_string (pid, n, _) = Printf.sprintf "%s_%s" pid n in 
  let to_string_var (pid, n, v) =
    Printf.sprintf "%s_%s=%s" pid n (Data.val_to_string string_of_float v)
  in
  let pot = match List.assoc_opt "potential" e.data with
    | Some F f -> Printf.sprintf " => potential %.1f" f
    | _ -> ""
  in
  let leg = match List.assoc_opt "legitimate" e.data with
    | Some B true -> Printf.sprintf "The current configuration is legitimate\n" 
    | Some B false -> ""
    | _ -> ""
  in
  let silent = if enab = [] then "The current configuration is silent\n"  else "" in
  let vars = List.rev vars in
  (* restore_round_nb e; *)
  (if short then
     Printf.sprintf "[round %i, step %i%s] %s %s\n" !round_st_ref.cpt e.step pot
       (if e.step <> e.nb then (":" ^ (string_of_int e.nb)) else "")
       (String.concat " " (List.map to_string_var vars))
   else
     Printf.sprintf "Round %i - Step %i%s%s\n%s%s\n%s%s" !round_st_ref.cpt e.step
       (if e.step <> e.nb then (" - Event " ^ (string_of_int e.nb)) else "") pot
       (RdbgStdLib.string_of_event e) 
       (if vars = [] then "" else
          ("Active node states: "^(String.concat " " (List.map to_string_var vars))))
       leg silent
  ) 
  
let print_sasa_event short e =
  if e.kind <> Ltop then print_event e else
    (
      print_string (str_of_sasa_event short e);
      flush stdout
    )

let _ =
  del_hook "print_event";
  add_hook "print_event" (print_sasa_event true)

(* shortcuts to the default and sasa event printers *)
let pe () = print_event !e;;
let spe () = print_sasa_event true !e;;
let spel () = print_sasa_event false !e;;

(**********************************************************************)
(* Goto to the next Lustre oracle violation *)
let goto_next_false_oracle e =
  next_cond e (fun e -> e.kind = Exit && e.lang = "lustre" &&
                          List.mem ("ok", Bool) e.outputs  &&
                            not (vb "ok" e))
  
let viol_string () =
  if args.oracles <> [] then (
    e:=goto_next_false_oracle !e; !dot_view (); "An oracle has been violated. Cf the .rif file"
  ) else (
    "No oracle is set."
  )
;;
let viol () = Printf.printf "%s\n%!" (viol_string ())

let _ = add_doc_entry
          "viol" "unit -> unit" "Move forward until the oracle is violated"
          "sasa" "sasa-rdbg-cmds.ml"

(**********************************************************************)
(* Potential  *)

let potential_opt () = List.assoc_opt "potential" !e.data

let potential () =
  match potential_opt () with
  | Some F f -> Some f
  | _ -> None

let _ = add_doc_entry
          "potential" "unit -> float option" "returns the current potential if available"
          "sasa" "sasa-rdbg-cmds.ml"



(**********************************************************************)
(* Perform the checkpointing at rounds! *)
let _ = check_ref := fun e -> e.nb = 1 || round e;;

(**********************************************************************)
let _ = 
  add_doc_entry "d"
    "unit -> unit" "update the current network with dot" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ne"
    "unit -> unit" "update the current network with neato" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "tw"
    "unit -> unit" "update the current network with twopi" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ci"
    "unit -> unit" "update the current network with circo" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "fd"
    "unit -> unit" "update the current network with fdp" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "sf"
    "unit -> unit" "update the current network with sfdp" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "pa"
    "unit -> unit" "update the current network with patchwork" "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "os"
    "unit -> unit" "update the current network with osage" "sasa" "sasa-rdbg-cmds.ml";
  
  add_doc_entry "sd" "unit -> unit"
    "go to the next step and update the network with one of the GraphViz tools" "sasa"
    "sasa-rdbg-cmds.ml";
  add_doc_entry "bd" "unit -> unit" "go to the previous step and update the network"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "nr" "unit -> unit" "go to the next round and update the network"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "pr" "unit -> unit" "go to the previous round and update the network"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "d_par" "unit -> unit" "cf d (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ne_par" "unit -> unit" "cf ne (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "tw_par" "unit -> unit" "cf tw (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "ci_par" "unit -> unit" "cf ci (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "fd_par" "unit -> unit" "cf fd (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "sf_par" "unit -> unit" "cf sf (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "pa_par" "unit -> unit" "cf pa (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  add_doc_entry "os_par" "unit -> unit" "cf os (for topology with a parent field)"
    "sasa" "sasa-rdbg-cmds.ml";
  ()

let l () =
  l();
  print_string (
      "
       - sasa commands
       + Display network with GraphViz tools
       d: "^(RdbgMain.doc_msg "d")^"
       ne: "^(RdbgMain.doc_msg "ne")^"
       tw: "^(RdbgMain.doc_msg "tw")^"
       ci: "^(RdbgMain.doc_msg "ci")^"
       fd: "^(RdbgMain.doc_msg "fd")^"
       sf: "^(RdbgMain.doc_msg "sf")^"
       pa: "^(RdbgMain.doc_msg "pa")^"
       os: "^(RdbgMain.doc_msg "os")^"
                                      
       nb: for algorithms that have a field named 'par', you can try 
           one of the following (which only draw the parent arcs)
       d_par: "^(RdbgMain.doc_msg "d")^" (parent arcs only)
       ne_par: "^(RdbgMain.doc_msg "ne")^" (parent arcs only)
       tw_par: "^(RdbgMain.doc_msg "tw")^" (parent arcs only)
       ci_par: "^(RdbgMain.doc_msg "ci")^" (parent arcs only)
       fd_par: "^(RdbgMain.doc_msg "fd")^" (parent arcs only)
       sf_par: "^(RdbgMain.doc_msg "sf")^" (parent arcs only)
       pa_par: "^(RdbgMain.doc_msg "pa")^" (parent arcs only)
       os_par: "^(RdbgMain.doc_msg "os")^" (parent arcs only)
       
       + Moving commands [*]
         sd: "^(RdbgMain.doc_msg "sd")^"
         nd: "^(RdbgMain.doc_msg "nd")^"
         bd: "^(RdbgMain.doc_msg "bd")^"
         nr: "^(RdbgMain.doc_msg "nr")^"
         pr: "^(RdbgMain.doc_msg "pr")^"
      [*] in order to change the current graph drawing engine, you can use
          the dot_view reference as follows:
           dot_view := ci
");;


(**********************************************************************)
(* ok, let's start the debugging session! *)

let pdf_viewer =
  if Sys.command "which zathura" = 0 then "zathura" else
    if Sys.command "which xpdf" = 0 then "xpdf" else
      if Sys.command "which acroread" = 0 then "acroread" else
        if Sys.command "which evince" = 0 then "evince" else (
          Printf.printf "Warning: no pdf viewer is found to visualize %s\n%!" dotfile;
          "ls"
        )
              
let graph_view () = 
  !dot_view ();
  let cmd = Printf.sprintf "%s sasa-%s.pdf&" pdf_viewer dotfile in
  Printf.printf "%s\n!" cmd;
  ignore(Sys.command cmd)

;;
(* let _ = graph_view () *)
let gv = graph_view

let _ = 
  add_doc_entry "gv" "unit -> unit"
    "launch a pdf viewer of the current network

     in order to change the current graph drawing engine, you can use
     the dot_view reference as follows:
     (rdbg)  dot_view := ci;;
     or
     (rdbg)  dot_view := fd;;

     In order to get the list of  graph drawing engine:
     (rdbg)   h sasa
     "
    "sasa" "sasa-rdbg-cmds.ml";
